# core/api.py

"""Build real URL for api request"""

from rcdl.core.models import Creator


class URL:
    """Build real URL for api request"""

    DOMAINS_BASE_URL = {
        "coomer": "https://coomer.st/api/v1/",
        "kemono": "https://kemono.cr/api/v1/",
    }

    @staticmethod
    def get_base_url(domain: str) -> str:
        """Return https://domain.com"""
        if domain not in URL.DOMAINS_BASE_URL:
            raise KeyError(f"{domain} not in known domains urls")
        return URL.DOMAINS_BASE_URL[domain]

    @staticmethod
    def get_post_revision(creator: Creator, post_id) -> str:
        """Return post revision url"""
        return (
            f"{URL.get_base_url(creator.domain)}{creator.service}"
            f"/user/{creator.creator_id}/post/{post_id}/revisions"
        )

    @staticmethod
    def get_headers() -> dict:
        """Return necessary request header for successful request"""
        return {
            "User-Agent": (
                "Mozilla/5.0 (Windows NT 10.0; Win64; x64) "
                "AppleWebKit/537.36 (KHTML, like Gecko) "
                "Chrome/117.0 Safari/537.36"
            ),
            "Accept": "text/css",
        }

    @staticmethod
    def get_url_from_file(domain: str, path_url: str):
        """Add path_url to based domain url"""
        if domain == "coomer":
            return f"https://coomer.st{path_url}"
        if domain == "kemono":
            return f"https://kemono.cr{path_url}"

        raise ValueError(
            f"Domain {domain} is not an accepted value/does not exist. "
            f"Please check your creators.json file"
        )

    @staticmethod
    def add_params(url: str, params: dict):
        """Create all parameters string (key=params&key=...)"""
        url += "?"
        for key in params:
            url += f"{key}={params[key]}&"
        return url[:-1]

    @staticmethod
    def get_creator_post_wo_param(creator: Creator) -> str:
        """Get creator post without parameters"""
        return (
            f"{URL.get_base_url(creator.domain)}{creator.service}"
            f"/user/{creator.creator_id}/posts"
        )

    @staticmethod
    def get_posts_page_url_wo_param():
        """Get posts page without parameters -> use in tag search"""
        domain = URL.DOMAINS_BASE_URL["coomer"]
        return f"{domain}posts"



# interface/ui.py

import logging
import click
from rich.console import Console, Group
from rich.table import Table
from rich.progress import (
    Progress,
    SpinnerColumn,
    BarColumn,
    TextColumn,
    TimeRemainingColumn,
)
from rich import box
from rich.live import Live
from rich.text import Text
from rcdl.core.models import Creator, FusedStatus, Status


class UI:
    console = Console()
    logger = logging.getLogger()

    _video_progress_text: Text | None = None
    _concat_progress_text: Text | None = None
    _live: Live | None = None

    @staticmethod
    def _log_to_file(log_level, msg: str):
        log_level(msg)

    @classmethod
    def success(cls, msg: str):
        """Print success msg"""
        cls.console.print(f"[green]{msg}[/]")

    @classmethod
    def info(cls, msg: str):
        """Print & log info msg"""
        cls.console.print(msg)
        cls._log_to_file(cls.logger.info, msg)

    @classmethod
    def debug(cls, msg: str):
        """Log debug msg"""
        # cls.console.print(f"[dim]{msg}[/]")
        cls._log_to_file(cls.logger.debug, msg)

    @classmethod
    def warning(cls, msg: str):
        """Print & log warning msg"""
        cls.console.print(f"[yellow]{msg}[/]")
        cls._log_to_file(cls.logger.debug, msg)

    @classmethod
    def error(cls, msg: str):
        """Print & log error msg"""
        cls.console.print(f"[red]{msg}[/]")
        cls._log_to_file(cls.logger.debug, msg)

    @classmethod
    def table_creators(cls, creators: list[Creator]):
        """Print to cli a table with all creators in creators.txt. Format is Creator ID | Service"""
        table = Table(title="Creators", box=box.MINIMAL, show_lines=True)
        table.add_column("Creators ID")
        table.add_column("Service")
        for creator in creators:
            table.add_row(creator.creator_id, creator.service)
        cls.console.print(table)

    @classmethod
    def progress_posts_fetcher(cls, max_pages: int):
        progress = Progress(
            SpinnerColumn(),
            TextColumn("[progress.description]{task.description}"),
            BarColumn(),
            console=cls.console,
            transient=False,  # remove progress bar after finish
        )
        return progress

    @classmethod
    def video_progress(cls, total: int):
        """Create video download progress output"""
        progress = Progress(
            SpinnerColumn(),
            TextColumn("[bold cyan]{task.description}"),
            BarColumn(),
            TextColumn("{task.completed}/{task.total}"),
            TimeRemainingColumn(),
            console=cls.console,
            transient=False,  # remove the bar after completion
        )

        cls._video_progress_text = Text("Waiting...", style="Cyan")
        group = Group(progress, cls._video_progress_text)
        cls._live = Live(group, console=cls.console)
        cls._live.__enter__()

        task = progress.add_task("Downloading videos", total=total)
        return progress, task

    @classmethod
    def set_current_video_progress(cls, creator_info: str, filename: str):
        """Update video download output
        args:
            creator_info: str =  'creator_id@(service)'
            filename: str = video.relative_path
        """
        if cls._video_progress_text is None:
            return
        cls._video_progress_text.plain = ""
        cls._video_progress_text.append(f"{creator_info} -> ", style="Cyan")
        cls._video_progress_text.append(filename, style="green")

    @classmethod
    def close_video_progress(cls):
        """Close video progress"""
        if cls._live:
            cls._live.__exit__(None, None, None)
            cls._live = None

    @classmethod
    def concat_progress(cls, total: int):
        """Create concat progress bat"""
        progress = Progress(
            SpinnerColumn(),
            TextColumn("[bold cyan]{task.description}"),
            BarColumn(),
            TextColumn("{task.completed}/{task.total}"),
            TimeRemainingColumn(),
            console=cls.console,
            transient=False,  # remove the bar after completion
        )

        cls._concat_progress_text = Text("Waiting...", style="Cyan")
        group = Group(progress, cls._concat_progress_text)
        cls._live = Live(group, console=cls.console)
        cls._live.__enter__()

        task = progress.add_task("Concatenating videos", total=total)
        return progress, task

    @classmethod
    def set_current_concat_progress(cls, msg: str, filename: str):
        """Update video download output
        args:
            creator_info: str =  'creator_id@(service)'
            filename: str = video.relative_path
        """
        if cls._concat_progress_text is None:
            return
        cls._concat_progress_text.plain = ""
        cls._concat_progress_text.append(f"{msg} -> ", style="Cyan")
        cls._concat_progress_text.append(filename, style="green")

    @classmethod
    def close_concat_progress(cls):
        """Close video progress"""
        if cls._live:
            cls._live.__exit__(None, None, None)
            cls._live = None

    @classmethod
    def progress_total_concat(cls):
        progress = Progress(
            SpinnerColumn(),
            TextColumn("[progress.description]{task.description}"),
            BarColumn(),
            TextColumn("{task.completed}/{task.total}"),
            console=cls.console,
            transient=False,  # remove progress bar after finish
        )
        return progress


def print_db_info(info: dict):
    click.echo("--- TABLES ---")
    click.echo("Posts:")
    click.echo(f"\t{info['posts']} total")
    click.echo("FusedMedias:")
    for status in FusedStatus:
        click.echo(f"\t{info['fuses'][status]} {status}")
    click.echo("Medias:")
    for status in Status:
        click.echo(f"\t{info['medias'][status]} {status}")



# gui/video_manager.py

import os

import streamlit as st

from rcdl.core.config import Config
from rcdl.core.models import Status, Media
from rcdl.core.db import DB


previous_statuses = {}


def set_status(media: Media, status: Status):
    key = media.post_id + media.url
    previous_statuses[key] = media.status
    media.status = status
    with DB() as db:
        db.update_media(media)
    print(f"Set {media.post_id} to {status.value}")

    for m in st.session_state.medias:
        if m.post_id == media.post_id and m.url == media.url:
            m.status = status
            break


def video_manager():
    st.title("Video Manager")

    # load db
    if "medias" not in st.session_state:
        with DB() as db:
            st.session_state.medias = db.query_medias_by_status_sorted(
                [Status.DOWNLOADED, Status.OPTIMIZED],
                sort_by="file_size",
                ascending=True,
            )

    medias = st.session_state.medias
    if not medias:
        st.info("No media found")
        return

    # session state
    if "media_index" not in st.session_state:
        st.session_state.media_index = 0

    idx = st.session_state.media_index
    media = medias[idx]

    # media info
    st.subheader(f"Media {idx + 1} / {len(medias)}")

    with DB() as db:
        post = db.query_post_by_id(media.post_id)
        if post is None:
            st.info("No matching post found")
            return

    col_video, col_info = st.columns([1, 2])
    with col_info:
        col1, col2 = st.columns(2)
        with col1:
            st.write("**Post ID:**", media.post_id)
            st.write("**Service:**", media.service)
            st.write("**User:**", post.user)
            st.write("**Duration:**", media.duration, "s")
            st.write("**Sequence:**", media.sequence)
            st.write("**Size:**", round(media.file_size / (1024 * 1024), 1), "MB")
            st.write("**Status:**", media.status)
            key = media.post_id + media.url
            if key in previous_statuses:
                st.write("**PREV STATUS:**", previous_statuses[key])
            st.write("**Path:**", media.file_path)
            st.write("**Created at**:", media.created_at[0:16])

        with col2:
            # controls
            c1, c2, c3 = st.columns(3)
            with c1:
                if st.button("⏮ Prev", disabled=idx == 0):
                    st.session_state.media_index -= 1
                    st.rerun()
            with c2:
                if st.button("⏭ Next", disabled=idx >= len(medias) - 1):
                    st.session_state.media_index += 1
                    st.rerun()
            with c3:
                if st.button("Remove"):
                    set_status(media, Status.TO_BE_DELETED)
                    st.rerun()
                if st.button("Not Remove"):
                    key = media.post_id + media.url
                    if key in previous_statuses:
                        set_status(media, previous_statuses[key])
                    else:
                        print("Not in previous status")
                    st.rerun()

    # video player
    full_path = os.path.join(Config.creator_folder(post.user), media.file_path)
    if os.path.exists(full_path):
        with col_video:
            with st.container():
                st.video(
                    full_path,
                    autoplay=True,
                    loop=True,
                )
    else:
        st.error(f"Video file {full_path} not found on disk")



# core/fuse.py

"""Handle merging videos from a same post"""

import os
import subprocess

from rcdl.core.config import Config
from rcdl.core.db import DB
from rcdl.core.models import FusedStatus, Status, FusedMedia, Media, Post
from rcdl.interface.ui import UI
import rcdl.core.downloader_subprocess as dls
from rcdl.utils import get_media_metadata, get_date_now


def update_db(fuse: FusedMedia, medias: list[Media], user: str, result):
    """Update DB depending on subprocess result (SUCESS/FAILURE)"""
    if result == 0:
        path = os.path.join(Config.creator_folder(user), fuse.file_path)
        duration, file_size, checksum = get_media_metadata(path)
        fuse.duration = duration
        fuse.status = FusedStatus.FUSED
        fuse.checksum = checksum
        fuse.created_at = get_date_now()
        fuse.file_size = file_size
        for media in medias:
            media.status = Status.FUSED
    else:
        fuse.fail_count += 1
    with DB() as db:
        db.update_fuse(fuse)
        for media in medias:
            db.update_media(media)


def get_medias_and_post(
    post_id: str, total_parts: int
) -> tuple[None, None] | tuple[list[Media], Post]:
    """Get medias and post related to a fuse group.
    Return a list[Media] and a Post
    Handle Errors, return None, None"""
    # get associated post
    with DB() as db:
        post = db.query_post_by_id(post_id)
    if post is None:
        UI.error(f"Could not match fuses post id {post_id} to a post in post tables")
        return None, None

    # get all videos of a post
    with DB() as db:
        medias = db.query_media_by_post_id(post_id)

    # check number of media in db match total part expected in fused media
    if len(medias) != total_parts:
        UI.error(f"Found {len(medias)} videos part. Expected {total_parts}")
        return None, None

    # check all video are downloaded
    allowed_status = [Status.DOWNLOADED, Status.OPTIMIZED]
    if Config.DEBUG:
        allowed_status.append(Status.FUSED)
    ok = True
    for media in medias:
        if media.status not in allowed_status:
            ok = False
            break
    if not ok:
        return None, None

    # sort medias list
    sorted_medias = sorted(medias, key=lambda m: m.sequence)
    return sorted_medias, post


def fuse_medias():
    """Fuse all media part of a fuse group with status PENDING in DB fuses"""
    # get all fused media
    with DB() as db:
        fuses = db.query_fuses_by_status(FusedStatus.PENDING)
    if Config.DEBUG:
        with DB() as db:
            ok_fuses = db.query_fuses_by_status(FusedStatus.FUSED)
        fuses.extend(ok_fuses)

    with UI.progress_total_concat() as progress:
        task = progress.add_task("Total concat", total=len(fuses))

        for fm in fuses:
            medias, post = get_medias_and_post(fm.id, fm.total_parts)
            if medias is None or post is None:
                progress.advance(task)
                continue

            # concat medias
            result = 1
            try:
                result = dls.ffmpeg_concat(medias, post)
            except (OSError, subprocess.SubprocessError, ValueError) as e:
                UI.error(f"Failed to concat video (id:{post.id}) due to: {e}")

            # update db
            update_db(fm, medias, post.user, result)

            progress.update(
                task, advance=1, description=f"Concated media for id {fm.id}"
            )

            # remove part file
            for media in medias:
                media_full_path = os.path.join(
                    Config.creator_folder(post.user), media.file_path
                )
                try:
                    if Config.DEBUG:
                        UI.info(f"Skipped {media_full_path} removal")
                        continue
                    os.remove(media_full_path)
                    UI.info(f"Removed file {media_full_path}")
                except (FileNotFoundError, PermissionError) as e:
                    UI.error(
                        f"FileNotFound/PermissionError: Failed to "
                        f"remove media {media_full_path} due to: {e}"
                    )
                except OSError as e:
                    UI.error(f"Failed to remove media {media_full_path} due to: {e}")



# core/models.py

"""Hold all dataclass models and enums"""

from dataclasses import dataclass
from enum import Enum


class Status(Enum):
    """Status for media"""

    PENDING = "pending"  # to be downloaded
    DOWNLOADED = "downloaded"  # video has been downloaded
    FUSED = "fused"  # video has been fused, and impliitly removed
    TO_BE_DELETED = "to_be_delete"  # video has been marked for delete
    DELETED = "deleted"  # video has been deleted
    OPTIMIZED = "optimized"  # video has been optimized (reduce file size)


class FusedStatus(Enum):
    """Status for fused group"""

    PENDING = "pending"
    FUSED = "fused"


@dataclass
class Post:
    """Post model that shadow post dict response of request
    Partially used in posts db (check db_queries.py)
    """

    id: str
    user: str
    service: str
    domain: str
    title: str
    substring: str
    published: str
    file: dict
    attachments: list
    json_hash: str
    raw_json: str
    fetched_at: str


@dataclass
class Media:
    """Media model: use in medias DB"""

    post_id: str
    service: str
    url: str
    duration: float
    sequence: int
    status: Status
    checksum: str
    file_path: str
    created_at: str
    updated_at: str
    file_size: int
    fail_count: int = 0


@dataclass
class FusedMedia:
    """Fuses group models.
    Used in fuses db."""

    id: str
    duration: int
    total_parts: int
    status: FusedStatus
    checksum: str
    file_path: str
    created_at: str
    updated_at: str
    file_size: int
    fail_count: int = 0


@dataclass
class Creator:
    """Creator model. Status is depreciated"""

    creator_id: str
    service: str
    domain: str



# core/downloader_subprocess.py

"""
Handle all subprocess call to external tool (yt-dlp, ffmpeg, ...)
"""

import subprocess
import logging
from pathlib import Path
import os


from rcdl.interface.ui import UI
from rcdl.core import parser
from rcdl.core.models import Media, Post
from rcdl.core.config import Config


def ytdlp_clear_cache():
    """Clear yt-dlp cache"""
    cmd = ["yt-dlp", "--rm-cache-dir"]
    subprocess.run(cmd, check=False)


def kill_aria2c():
    """Kill all aria2c process"""
    cmd = ["pkill", "-f", "aria2c"]
    subprocess.run(cmd, check=False)


def ytdlp_subprocess(
    url: str,
    filepath: Path | str,
):
    """Call yt-dlp in a subprocess to download a video"""

    cmd = [
        "yt-dlp",
        "-q",
        "--progress",
        url,
        "-o",
        filepath,
        "--external-downloader",
        "aria2c",
    ]

    logging.info("CMD: %s", " ".join(cmd))

    result = subprocess.run(cmd, capture_output=True, text=True, check=False)
    if result.returncode != 0:
        logging.error("yt-dlp failed to dl vid: %s", result.stderr)

    return result.returncode


def ffprobe_get_duration(path: Path) -> int | None:
    """Get duration of a video in seconds with ffprobe
    Return an int or None if command failed"""
    cmd = [
        "ffprobe",
        "-v",
        "error",
        "-show_entries",
        "format=duration",
        "-of",
        "default=noprint_wrappers=1:nokey=1",
        str(path),
    ]
    try:
        result = subprocess.run(
            cmd,
            stdout=subprocess.PIPE,
            stderr=subprocess.DEVNULL,
            text=True,
            check=True,
        )
        return int(float(result.stdout.strip()))
    except subprocess.CalledProcessError as e:
        UI.error(f"Failed to use ffprobe on {path} due to {e}")
        return None
    except (AttributeError, ValueError, OverflowError) as e:
        UI.error(f"Failed to parse duration result of {path} due to {e}")
        return None


def get_max_width_height(medias: list[Media], post: Post) -> tuple[int, int]:
    """Get width and height of all media in list. Return max within video found and config"""

    cmd = [
        "ffprobe",
        "-v",
        "error",
        "-select_streams",
        "v:0",
        "-show_entries",
        "stream=width,height",
        "-of",
        "csv=p=0",
    ]
    width = 0
    height = 0
    max_width = 1920
    max_height = 1080
    for m in medias:
        path = os.path.join(Config.creator_folder(post.user), m.file_path)
        full_cmd = cmd + [path]

        try:
            result = subprocess.run(
                full_cmd, capture_output=True, text=True, check=True
            )
            w_str, h_str = result.stdout.strip().split(",")

            width = min(int(w_str), max_width)
            height = min(int(h_str), max_height)
        except subprocess.CalledProcessError as e:
            UI.error(f"Fail to use ffprobe to get width, height on {path} due to {e}")
        except (AttributeError, ValueError, OverflowError) as e:
            UI.error(f"Failed to parse duration for {path} due to {e}")
    return (width, height)


def get_total_duration(medias: list[Media], post: Post) -> int:
    """Get total duration in ms of all medias in list"""

    def _get_duration(path: str) -> int:
        """Get video duration in ms"""

        cmd = [
            "ffprobe",
            "-v",
            "error",
            "-select_streams",
            "v:0",
            "-show_entries",
            "format=duration",
            "-of",
            "default=noprint_wrappers=1:nokey=1",
            path,
        ]

        result = subprocess.run(cmd, capture_output=True, text=True, check=True)
        return int(float(result.stdout.strip()) * 1000)

    duration = 0
    for m in medias:
        path = os.path.join(Config.creator_folder(post.user), m.file_path)
        duration += _get_duration(path)
    return duration


def ffmpeg_concat_build_command(medias: list[Media], post: Post) -> dict:
    """Build the ffmpeg concat command"""

    width, height = get_max_width_height(medias, post)
    logging.info("Found (%s, %s) (width, height) for this group.", width, height)
    if width == 0:
        width = Config.MAX_WIDTH
    if height == 0:
        height = Config.MAX_HEIGHT

    # output path
    output_filename = parser.get_filename_fuse(post)
    output_path = os.path.join(Config.creator_folder(post.user), output_filename)

    # build cmd
    cmd = ["ffmpeg", "-y", "-progress", "pipe:2", "-nostats"]

    # inputs
    for media in medias:
        input_path = os.path.join(Config.creator_folder(post.user), media.file_path)
        cmd.extend(["-i", input_path])

    # filter complex
    filter_lines = []
    for idx in range(len(medias)):
        filter_lines.append(
            f"[{idx}:v]"
            f"scale={width}:{height}:force_original_aspect_ratio=decrease,"
            f"pad={width}:{height}:(ow-iw)/2:(oh-ih)/2,"
            f"fps={Config.FPS},setsar=1"
            f"[v{idx}]"
        )

    # concat inputs
    concat = []
    for idx in range(len(medias)):
        concat.append(f"[v{idx}][{idx}:a]")

    filter_lines.append(f"{''.join(concat)}concat=n={len(medias)}:v=1:a=1[outv][outa]")
    filter_complex = ";".join(filter_lines)

    cmd.extend(
        [
            "-filter_complex",
            filter_complex,
            "-map",
            "[outv]",
            "-map",
            "[outa]",
            "-c:v",
            "libx264",
            "-preset",
            Config.PRESET,
            "-threads",
            str(Config.THREADS),
            "-c:a",
            "aac",
            "-movflags",
            "+faststart",
            output_path,
        ]
    )

    return {"cmd": cmd, "output_path": output_path}


def ffmpeg_concat(medias: list[Media], post: Post):
    """Run ffmpeg concat command to merge video together"""

    command_builder = ffmpeg_concat_build_command(medias, post)
    cmd = command_builder["cmd"]
    output_path = command_builder["output_path"]

    logging.info("CMD: %s", " ".join(cmd))

    ffmpeg_log = Config.CACHE_DIR / "ffmpeg.log"
    with open(ffmpeg_log, "w", encoding="utf-8") as log_file:
        print(cmd, file=log_file)
        # run cmd
        process = subprocess.Popen(
            cmd,
            stdout=subprocess.DEVNULL,
            stderr=subprocess.PIPE,
            text=True,
            bufsize=1,
        )

        assert process.stderr is not None
        total_duration = get_total_duration(medias, post)
        progress, task = UI.concat_progress(total=total_duration)
        last_progress = 0
        UI.set_current_concat_progress(f"{medias[0].file_path}", output_path)

        for line in process.stderr:
            line = line.strip()
            if not line:
                continue

            print(line, file=log_file)

            progres_key = "out_time_ms"
            if line.startswith(progres_key):
                current_progress_str = line.replace(f"{progres_key}=", "").strip()
                try:
                    current_progress_us = int(current_progress_str)
                    current_progress_ms = current_progress_us // 1000
                    delta = current_progress_ms - last_progress
                    progress.advance(task, advance=delta)
                    last_progress = current_progress_ms
                except ValueError as e:
                    logging.warning(
                        "Skipping invalid progress line: %r (%s)",
                        current_progress_str,
                        e,
                    )
                except Exception as e:
                    UI.error(f"Unexpected error while updating progress: {e}")

        process.wait()
        UI.close_concat_progress()

    UI.debug(f"Result: {process.returncode}")
    if process.returncode != 0:
        UI.error(f"Failed to concat videos. See ffmpeg log file {ffmpeg_log}")
        with open(ffmpeg_log, "r", encoding="utf-8") as f:
            lines = f.read()
        logging.warning("---FFMPEG LOG---")
        logging.warning(lines)
        logging.warning("---END FFMPEG LOG---")
        return process.returncode

    return 0


def handbrake_optimized(media: Media, user: str):
    """Optimize video size with handbrake software"""

    handbrake_process = Config.HANDBRAKE_RUN_CMD.split(" ")

    folder_path = Config.creator_folder(user)
    video_path = os.path.join(folder_path, media.file_path)

    output_path = video_path + ".opti.mp4"

    cmd = ["-i", video_path, "-o", output_path, "--preset", "HQ 1080p30 Surround"]

    full_cmd = handbrake_process + cmd
    UI.info(f"Running cmd '{full_cmd}'")

    result = subprocess.run(full_cmd, check=False)

    if result.returncode == 0:
        UI.info("Return code: 0")
    else:
        UI.error(f"Return code: {result.returncode}")

    return result.returncode



# core/db.py

"""
Handle SQL Database
"""

import sqlite3

from rcdl.core import adapters
from rcdl.core import db_queries as queries
from rcdl.core.config import Config
from rcdl.core.models import Post, Media, Status, FusedMedia, FusedStatus
from rcdl.utils import get_date_now

from rcdl.interface.ui import UI


class DB:
    """Handle all sqlite database command"""

    def __init__(self):
        self.conn = sqlite3.connect(Config.DB_PATH)
        self.conn.row_factory = sqlite3.Row

    def __enter__(self):
        """necessary to use with openDB()"""
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        """necessary to use with openDB()"""
        self.close()

    def close(self):
        """Properly close database"""
        self.conn.close()

    def init_database(self):
        """Create tables (posts, fuses, medias) if they dont exist"""
        self.conn.execute(queries.CREATE_POSTS_TABLE)
        self.conn.execute(queries.CREATE_MEDIAS_TABLE)
        self.conn.execute(queries.CREATE_FUSE_TABLE)

        self.conn.commit()

    def get_nb_per_status(self):
        """Return an info dict per tables with number of entry per status
        info['tables1']['status1'] = X
        ...
        """
        info = {}
        info["medias"] = {}
        info["fuses"] = {}
        info["posts"] = 0
        for status in Status:
            info["medias"][status] = len(self.query_media_by_status(status))
        for status in FusedStatus:
            info["fuses"][status] = len(self.query_fuses_by_status(status))

        cur = self.conn.execute(("SELECT COUNT(*) AS count FROM posts"))
        row = cur.fetchone()
        info["post"] = row["count"] if row else 0
        return info

    def query_post_by_id(self, _id: str) -> Post | None:
        """Get a post from his post id"""
        row = self.conn.execute(queries.QUERY_POST_ID, (_id,)).fetchone()

        UI.debug(f"{queries.QUERY_POST_ID} {_id} returned {row}")

        if row is None:
            return None

        return adapters.row_to_post(row)

    def query_post_by_user(self, user: str) -> list[Post]:
        """Get all posts of a user"""
        cur = self.conn.cursor()
        cur.execute(queries.QUERY_POST_USER, (user,))
        rows = cur.fetchall()

        UI.debug(f"{queries.QUERY_POST_USER} {user} returned {len(rows)} results")

        return adapters.rows_to_posts(rows)

    def query_media_by_status(self, status: Status) -> list[Media]:
        """Get all medias with specified status"""
        rows = self.conn.execute(queries.QUERY_MEDIA_STATUS, (status.value,)).fetchall()
        UI.debug(
            f"{queries.QUERY_MEDIA_STATUS} {status.value} returned {len(rows)} result"
        )

        return adapters.rows_to_medias(rows)

    def query_medias_by_status_sorted(
        self,
        status: Status | list[Status],
        sort_by: str | None = None,
        ascending: bool = True,
    ) -> list[Media]:
        """Get all medias with specified status (one or multiple)
        Return them sorted by column and asc or desc"""

        # validate sort column
        valid_columns = {
            "id",
            "post_id",
            "service",
            "url",
            "duration",
            "sequence",
            "status",
            "checksum",
            "file_path",
            "created_at",
            "updated_at",
            "file_size",
            "fail_count",
        }
        order_clause = ""
        if sort_by:
            if sort_by not in valid_columns:
                UI.error(f"Invalid sort column: {sort_by}")
            order_clause = f"ORDER BY {sort_by} {'ASC' if ascending else 'DESC'}"

        # status filter
        if isinstance(status, Status):
            status = [status]

        status_values = [s.value if isinstance(s, Status) else s for s in status]
        placeholders = ", ".join("?" for _ in status_values)

        sql = f"SELECT * FROM medias WHERE status IN ({placeholders}) {order_clause}"
        rows = self.conn.execute(sql, status_values).fetchall()

        UI.debug(
            f"Queried medias with status={status_values}, sorted by {sort_by}, ascending={ascending}, {len(rows)} results"
        )

        return adapters.rows_to_medias(rows)

    def query_media_by_post_id(self, _id: str) -> list[Media]:
        """Get all medias from the same post by post id"""
        rows = self.conn.execute(queries.QUERY_MEDIA_ID, (_id,)).fetchall()
        UI.debug(f"{queries.QUERY_MEDIA_ID} {_id} returned {len(rows)} result")
        return adapters.rows_to_medias(rows)

    def query_fuses_by_status(self, status: FusedStatus) -> list[FusedMedia]:
        """Get all fused_media with specified status"""
        rows = self.conn.execute(queries.QUERY_FUSES_STATUS, (status.value,)).fetchall()
        UI.debug(
            f"{queries.QUERY_FUSES_STATUS} {status.value} returned {len(rows)} result"
        )

        return adapters.rows_to_fuses(rows)

    def query_fuses_by_id(self, _id: str) -> FusedMedia | None:
        """Get a fuse group by its unique post id"""
        row = self.conn.execute(queries.QUERY_FUSES_ID, (_id,)).fetchone()
        UI.debug(f"{queries.QUERY_FUSES_ID} {_id} returned {row} result")
        return adapters.row_to_fused_media(row)

    def insert_posts(self, posts: list[Post] | Post):
        """Add post to DB if it does not already exist (UNIQUE post_id)"""
        if isinstance(posts, Post):
            posts = [posts]

        values = []
        for post in posts:
            values.append(
                (
                    post.id,
                    post.user,
                    post.service,
                    post.domain,
                    post.published,
                    post.json_hash,
                    post.raw_json,
                    post.fetched_at,
                )
            )

        with self.conn:
            self.conn.executemany(queries.INSERT_POST, values)

        inserted = self.conn.total_changes
        UI.debug(f"Inserted {inserted} new posts out of {len(posts)} total posts")

    def insert_medias(self, medias: list[Media] | Media):
        """Insert media into the db if it does not already exist (UNIQUE post_id, url)"""
        if isinstance(medias, Media):
            medias = [medias]

        values = []
        for media in medias:
            values.append(
                (
                    media.post_id,
                    media.service,
                    media.url,
                    media.duration,
                    media.sequence,
                    media.status.value,
                    media.checksum,
                    media.file_path,
                    media.created_at,
                    get_date_now(),
                    media.file_size,
                    media.fail_count,
                )
            )

        with self.conn:
            self.conn.executemany(queries.INSERT_MEDIA, values)

        inserted = self.conn.total_changes
        UI.debug(f"Inserted {inserted} new media out of {len(medias)} total medias")

    def update_media(self, media: Media):
        """Update media entry in the db. Found it by post_id & url, and update:
        - duration, file_size, checksum, status, create_at, updated_at, fail_count"""
        params = (
            media.duration,
            media.file_size,
            media.checksum,
            media.status.value,
            media.created_at,
            get_date_now(),
            media.fail_count,
            media.post_id,
            media.url,
        )
        with self.conn:
            self.conn.execute(queries.UPDATE_MEDIA, params)
            UI.debug(f"Updated media {media.post_id} / {media.url}")

    def insert_fused_media(self, fuses: list[FusedMedia] | FusedMedia):
        """Insert fused_media into the db if it does not already exist (UNIQUE post_id)"""
        if isinstance(fuses, FusedMedia):
            fuses = [fuses]

        values = []
        for fuse in fuses:
            values.append(
                (
                    fuse.id,
                    fuse.duration,
                    fuse.total_parts,
                    fuse.status.value,
                    fuse.checksum,
                    fuse.file_path,
                    fuse.created_at,
                    get_date_now(),
                    fuse.file_size,
                    fuse.fail_count,
                )
            )

        with self.conn:
            self.conn.executemany(queries.INSERT_FUSED_MEDIA, values)

        inserted = self.conn.total_changes
        UI.debug(
            f"Inserted {inserted} new fused_media out of {len(fuses)} total fused_media"
        )

    def update_fuse(self, fuse: FusedMedia):
        """Update fuse group: duration, status, checksum,
        created_at, updated_at, file_size, fail_count

        """
        params = (
            fuse.duration,
            fuse.status.value,
            fuse.checksum,
            fuse.created_at,
            get_date_now(),
            fuse.file_size,
            fuse.fail_count,
            fuse.id,
        )
        with self.conn:
            self.conn.execute(queries.UPDATE_FUSE, params)
            UI.debug(f"Updated fuse {fuse.id} / {fuse.file_path}")



# utils.py

"""Hold collection of useful functions"""

from datetime import datetime, timezone
import hashlib
import json
from pathlib import Path
import os
import time

import rcdl.core.downloader_subprocess as dls
from rcdl.core.config import Config
from rcdl.interface.ui import UI


def get_date_now() -> str:
    """Return an str of current datetime in isoformat"""
    return datetime.now(timezone.utc).isoformat()


def get_media_metadata(full_path: str) -> tuple[int, int, str]:
    """Get duration (if possible), file_size and checksum"""
    path = Path(full_path)
    if not os.path.exists(path):
        UI.error(f"{path} path should exist but does not.")

    file_size = path.stat().st_size
    checksum = get_media_hash(path)
    if checksum is None:
        checksum = ""
    duration = dls.ffprobe_get_duration(path)
    if duration is None:
        duration = 0
    return duration, file_size, checksum


def get_media_hash(path: Path, retries: int = Config.CHECKSUM_RETRY) -> str | None:
    """Return a hash of a video file"""
    for attempt in range(retries + 1):
        try:
            sha256_hash = hashlib.sha256()
            with path.open("rb") as f:
                for chunk in iter(lambda: f.read(8192), b""):
                    sha256_hash.update(chunk)
            checksum = sha256_hash.hexdigest()
            return checksum
        except OSError as e:
            if attempt >= retries:
                UI.error(
                    f"OSError: Failed to checksum {path} "
                    f"at attempt {attempt + 1} due to: {e}"
                )
                return None
            UI.warning(
                f"OSError: Failed to checksum {path} "
                f"at attempt {attempt + 1} due to: {e}"
            )
            time.sleep(1.0)


def get_json_hash(data: dict) -> tuple[str, str]:
    """Return a hash of a dict"""
    raw_json = json.dumps(data, sort_keys=True)
    json_hash = hashlib.sha256(raw_json.encode("utf-8")).hexdigest()
    return raw_json, json_hash


def bytes_to_mb(bytes: float | int):
    return round(bytes / (1024 * 1024), 1)


def clean_all(all: bool, partial: bool, cache: bool, medias_deleted: bool):
    """Remove partial file & external programm cache"""

    from rcdl.core.db import DB
    from rcdl.core.models import Status

    if all:
        partial = True
        cache = True
        medias_deleted = True

    # remove all partial file
    if partial:
        path = Config.BASE_DIR
        folders = os.listdir(path)

        for folder in folders:
            if folder.startswith("."):
                continue

            folder_full_path = os.path.join(path, folder)
            files = os.listdir(folder_full_path)

            for file in files:
                if (
                    file.endswith(".aria2")
                    or file.endswith(".part")
                    or file.endswith(".opti.mp4")
                ):
                    full_path = os.path.join(folder_full_path, file)
                    os.remove(full_path)
                    UI.info(f"Removed {full_path}")

    # cache
    if cache:
        # clear yt-dlp cache
        dls.ytdlp_clear_cache()
        UI.info("Cleared yt-dlp cache dir")

        # kill aria2c
        dls.kill_aria2c()
        UI.info("Kill aria2c")

    # medias with status deleted
    if medias_deleted:
        with DB() as db:
            medias = db.query_media_by_status(Status.TO_BE_DELETED)
        UI.info(f"Found {len(medias)} medias with DELETED status")

        if len(medias) == 0:
            return

        total_size = 0.0
        for media in medias:
            with DB() as db:
                post = db.query_post_by_id(media.post_id)
            if post is None:
                UI.warning(f"Could not match post id {media.post_id} to a post")
                continue

            path = os.path.join(Config.creator_folder(post.user), media.file_path)
            if not os.path.exists:
                UI.error(f"Path {path} should exist. Does not")
                continue

            try:
                os.remove(path)
                UI.info(f"Removed '{path}' ({bytes_to_mb(media.file_size)} MB)")
                total_size += media.file_size
            except Exception as e:
                UI.error(f"Failed to rm media {media.post_id}/{media.url} due to: {e}")
                continue

            with DB() as db:
                media.status = Status.DELETED
                db.update_media(media)

        UI.info(
            f"Removed a total of {bytes_to_mb(total_size)} MB,"
            f" {round(bytes_to_mb(total_size) / 100, 1)} GB"
        )



# gui/gui.py

import streamlit as st

from rcdl.gui.db_viewer import run_db_viewer
from rcdl.gui.video_manager import video_manager

st.markdown(
    """
    <style>
    /* Remove top padding */
    .block-container {
        padding-top: 1rem !important;
    }

    /* Optional: remove Streamlit header */
    header[data-testid="stHeader"] {
        display: none;
    }

    /* Optional: remove footer */
    footer {
        display: none;
    }
    </style>
    """,
    unsafe_allow_html=True,
)


def run_gui():
    """
    Launches the Streamlit GUI.
    This function can be called from a CLI command.
    """
    # Streamlit code
    st.set_page_config(page_title="RCDL", layout="wide")

    # Sidebar navigation
    page = st.sidebar.radio("Go to", ["Home", "Manage Videos", "View DB"])

    if page == "Home":
        st.header("Home Page")
        st.write("Develloped by - ritonun -")

    elif page == "Manage Videos":
        video_manager()

    elif page == "View DB":
        run_db_viewer()


if __name__ == "__main__":
    run_gui()
